from datetime import timedelta
from uuid import uuid4

from django.conf import settings
from django.utils.translation import ugettext_lazy as _

from micro_framework.jwt_auth.exceptions import TokenBackendError, TokenError
from micro_framework.jwt_auth.redis_token.token_manager import \
    RedisTokenManager
from micro_framework.jwt_auth.utils import (aware_utcnow, datetime_from_epoch,
                                            datetime_to_epoch, format_lazy)
from micro_framework.settings import api_settings


class Token:
    """
    A class which validates and wraps an existing JWT or can be used to build a
    new JWT.
    """
    token_type = None
    lifetime = None

    def __init__(self, token=None, verify=True):
        """
        !!!! IMPORTANT !!!! MUST raise a TokenError with a user-facing error
        message if the given token is invalid, expired, or otherwise not safe
        to use.
        """
        if self.token_type is None or self.lifetime is None:
            raise TokenError(_('Cannot create token with no type or lifetime'))

        self.token = token
        self.current_time = aware_utcnow()

        # Set up token
        if token is not None:
            # An encoded token was provided
            from .state import token_backend

            # Decode token
            try:
                self.payload = token_backend.decode(token, verify=verify)
            except TokenBackendError:
                raise TokenError(_('Token is invalid or expired'))

            if verify:
                self.verify()
        else:
            # New token.  Skip all the verification steps.
            self.payload = {api_settings.TOKEN_TYPE_CLAIM: self.token_type}

            # Set "exp" claim with default value
            self.set_exp(from_time=self.current_time, lifetime=self.lifetime)

            # Set "jti" claim
            self.set_jti()

    def __repr__(self):
        return repr(self.payload)

    def __getitem__(self, key):
        return self.payload[key]

    def __setitem__(self, key, value):
        self.payload[key] = value

    def __delitem__(self, key):
        del self.payload[key]

    def __contains__(self, key):
        return key in self.payload

    def get(self, key, default=None):
        return self.payload.get(key, default)

    def __str__(self):
        """
        Signs and returns a token as a base64 encoded string.
        """
        from micro_framework.jwt_auth.state import token_backend

        return token_backend.encode(self.payload)

    def verify(self):
        """
        Performs additional validation steps which were not performed when this
        token was decoded.  This method is part of the "public" API to indicate
        the intention that it may be overridden in subclasses.
        """
        # According to RFC 7519, the "exp" claim is OPTIONAL
        # (https://tools.ietf.org/html/rfc7519#section-4.1.4).  As a more
        # correct behavior for authorization tokens, we require an "exp"
        # claim.  We don't want any zombie tokens walking around.
        self.check_exp()

        # Ensure token id is present
        if api_settings.JTI_CLAIM not in self.payload:
            raise TokenError(_('Token has no id'))

        self.verify_token_type()

    def verify_token_type(self):
        """
        Ensures that the token type claim is present and has the correct value.
        """
        try:
            token_type = self.payload[api_settings.TOKEN_TYPE_CLAIM]
        except KeyError:
            raise TokenError(_('Token has no type'))

        if self.token_type != token_type:
            raise TokenError(_('Token has wrong type'))

    def set_jti(self):
        """
        Populates the configured jti claim of a token with a string where there
        is a negligible probability that the same string will be chosen at a
        later time.

        See here:
        https://tools.ietf.org/html/rfc7519#section-4.1.7
        """
        self.payload[api_settings.JTI_CLAIM] = uuid4().hex

    def set_exp(self, claim='exp', from_time=None, lifetime=None):
        """
        Updates the expiration time of a token.
        """
        if from_time is None:
            from_time = self.current_time

        if lifetime is None:
            lifetime = self.lifetime

        self.payload[claim] = datetime_to_epoch(from_time + lifetime)

    def check_exp(self, claim='exp', current_time=None):
        """
        Checks whether a timestamp value in the given claim has passed (since
        the given datetime value in `current_time`).  Raises a TokenError with
        a user-facing error message if so.
        """
        if current_time is None:
            current_time = self.current_time

        try:
            claim_value = self.payload[claim]
        except KeyError:
            raise TokenError(format_lazy(_("Token has no '{}' claim"), claim))

        claim_time = datetime_from_epoch(claim_value)
        if claim_time <= current_time:
            raise TokenError(format_lazy(_("Token '{}' claim has expired"), claim))

    @classmethod
    def for_user(cls, user, **kwargs):
        """
        Returns an authorization token for the given user that will be provided
        after authenticating the user's credentials.
        """
        user_id = getattr(user, api_settings.USER_ID_FIELD)
        if not isinstance(user_id, int):
            user_id = str(user_id)

        token = cls()
        token[api_settings.USER_ID_CLAIM] = user_id
        for key, value in kwargs.items():
            token[key] = value

        return token


class RedisTokenMixin:
    if 'micro_framework.jwt_auth.redis_token' in settings.INSTALLED_APPS:
        def verify(self, *args, **kwargs):
            self.check_redis_token()
            super().verify(*args, **kwargs)

        def check_redis_token(self):
            jti = self.payload[api_settings.JTI_CLAIM]
            redis_tokens = RedisTokenManager()
            try:
                redis_user_token = redis_tokens.get(jti)
            except Exception as ex:
                raise TokenError(_('Token is invalid'))
        
        @classmethod
        def for_user(cls, user, **kwargs):
            """
            Adds this token to the redis token list.
            """
            token = super().for_user(user, **kwargs)
            user_id = user.id
            jti = token[api_settings.JTI_CLAIM]
            redis_tokens = RedisTokenManager()
            try:
                redis_tokens.create(token_id=jti, user_id=user_id, payload=token.payload)
            except Exception as ex:
                raise TokenError(_("Something wrong with Token Storage Backend"))
            return token

class RefreshToken(RedisTokenMixin, Token):
    token_type = 'refresh'
    lifetime = api_settings.REFRESH_TOKEN_LIFETIME
    no_copy_claims = (
        api_settings.TOKEN_TYPE_CLAIM,
        'exp',

        # Both of these claims are included even though they may be the same.
        # It seems possible that a third party token might have a custom or
        # namespaced JTI claim as well as a default "jti" claim.  In that case,
        # we wouldn't want to copy either one.
        api_settings.JTI_CLAIM,
        'jti',
    )

    @property
    def access_token(self):
        """
        Returns an access token created from this refresh token.  Copies all
        claims present in this refresh token to the new access token except
        those claims listed in the `no_copy_claims` attribute.
        """
        access = AccessToken()

        # Use instantiation time of refresh token as relative timestamp for
        # access token "exp" claim.  This ensures that both a refresh and
        # access token expire relative to the same time if they are created as
        # a pair.
        access.set_exp(from_time=self.current_time)

        no_copy = self.no_copy_claims
        for claim, value in self.payload.items():
            if claim in no_copy:
                continue
            access[claim] = value

        return access


class AccessToken(Token):
    token_type = 'access'
    lifetime = api_settings.ACCESS_TOKEN_LIFETIME


class UntypedToken(Token):
    token_type = 'untyped'
    lifetime = timedelta(seconds=0)

    def verify_token_type(self):
        """
        Untyped tokens do not verify the "token_type" claim.  This is useful
        when performing general validation of a token's signature and other
        properties which do not relate to the token's intended use.
        """
        pass





from django.contrib.auth import get_user_model

from micro_framework.jwt_auth.backends import TokenBackend
from micro_framework.settings import api_settings

User = get_user_model()
token_backend = TokenBackend(api_settings.ALGORITHM, api_settings.SIGNING_KEY,
                             api_settings.VERIFYING_KEY, api_settings.AUDIENCE, api_settings.ISSUER)



# Generated by Django 3.0.1 on 2020-01-01 13:57

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Policy',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, unique=True)),
                ('rules', models.TextField()),
                ('notes', models.TextField(blank=True)),
            ],
            options={
                'verbose_name': 'Policy',
                'verbose_name_plural': 'Policies',
                'db_table': 'policies',
            },
        ),
    ]



from django.utils.functional import cached_property

from micro_framework.settings import api_settings
from micro_framework.store.redis_backend.client import redis_client
from micro_framework.store.redis_backend.utils import data_to_json, redis_to_data

# Cấu trúc tokens:<token_id>:<user_id>
TOKEN_KEY = "tokens:{}:{}"

def get_token_key(token_id, user_id):
    return TOKEN_KEY.format(token_id, user_id)

class RedisTokenManager:
    """
    Redis Token Manager
    """
    def create(self, token_id, user_id, payload):
        redis_backend = redis_client()
        redis_token_key = get_token_key(token_id, user_id)
        redis_timeout = None
        if api_settings.REDIS_EXPIRE_TOKEN:
            redis_timeout = int(api_settings.REFRESH_TOKEN_LIFETIME.total_seconds())
        redis_backend.set(
            redis_token_key,
            data_to_json(payload),
            ex=redis_timeout
        )
        return self.get(token_id)

    def list(self, user_id=None):
        redis_backend = redis_client()
        if not user_id:
            user_id = '*'
        list_token_by_user_id = redis_backend.keys(get_token_key('*', user_id))
        list_redis_token = []
        for redis_token_key in list_token_by_user_id:
            redis_token = RedisToken(redis_token_key, redis_backend.get(redis_token_key))
            list_redis_token.append(redis_token)
        return list_redis_token

    def get(self, token_id):
        redis_backend = redis_client()
        list_token_by_token_id = redis_backend.keys(get_token_key(token_id, '*'))
        if len(list_token_by_token_id) < 1:
            raise ValueError("Token does not exist")
        return RedisToken(list_token_by_token_id[0], redis_backend.get(list_token_by_token_id[0]))

    def delete(self, token_id):
        redis_backend = redis_client()
        list_token_by_token_id = redis_backend.keys(get_token_key(token_id, '*'))
        return redis_backend.delete(*list_token_by_token_id)

    def delete_all(self):
        redis_backend = redis_client()
        list_token_by_token_id = redis_backend.keys(get_token_key('*', '*'))
        if len(list_token_by_token_id) > 0:
            return redis_backend.delete(*list_token_by_token_id)
        return 0

class RedisToken:
    """
    Redis Token
    """
    def __init__(self, redis_key, payload):
        if isinstance(redis_key, bytes):
            redis_key = redis_key.decode("utf8")
        if isinstance(payload, bytes):
            payload = redis_to_data(payload)

        self.redis_key = redis_key
        self.payload = payload

    def __repr__(self):
        return "<RedisToken: %s>" % (self.redis_key or 'unknown')

    def __iter__(self):
        yield self.token_id, self.payload


    @cached_property
    def redis_backend(self):
        return redis_client()

    @property
    def token_id(self):
        return self.redis_key.split(':')[1]

    @property
    def user_id(self):
        return self.payload.get(api_settings.USER_ID_CLAIM, None)

    @property
    def time_to_live(self):
        return self.redis_backend.ttl(self.redis_key)

    def delete(self):
        return self.redis_backend.delete(self.redis_key)



from django.contrib.auth import authenticate
from django.utils.translation import ugettext_lazy as _
from rest_framework import exceptions, serializers

from micro_framework.helpers import (get_real_ip_from_header_meta,
                                     get_user_agent_from_header_meta)
from micro_framework.jwt_auth.exceptions import InvalidUser
from micro_framework.jwt_auth.state import User
from micro_framework.jwt_auth.tokens import RefreshToken, UntypedToken
from micro_framework.settings import api_settings


class PasswordField(serializers.CharField):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('style', {})

        kwargs['style']['input_type'] = 'password'
        kwargs['write_only'] = True

        super().__init__(*args, **kwargs)


class TokenObtainSerializer(serializers.Serializer):
    username_field = User.USERNAME_FIELD

    default_error_messages = {
        'no_active_account': _('No active account found with the given credentials')
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields[self.username_field] = serializers.CharField()
        self.fields['password'] = PasswordField()

    def validate(self, attrs):
        authenticate_kwargs = {
            self.username_field: attrs[self.username_field],
            'password': attrs['password'],
        }
        try:
            authenticate_kwargs['request'] = self.context['request']
        except KeyError:
            pass

        self.user = authenticate(**authenticate_kwargs)

        # Prior to Django 1.10, inactive users could be authenticated with the
        # default `ModelBackend`.  As of Django 1.10, the `ModelBackend`
        # prevents inactive users from authenticating.  App designers can still
        # allow inactive users to authenticate by opting for the new
        # `AllowAllUsersModelBackend`.  However, we explicitly prevent inactive
        # users from authenticating to enforce a reasonable policy and provide
        # sensible backwards compatibility with older Django versions.
        if self.user is None or not self.user.is_active:
            raise exceptions.AuthenticationFailed(
                self.error_messages['no_active_account'],
                'no_active_account',
            )

        return {}

    @classmethod
    def get_token(cls, user, **kwargs):
        raise NotImplementedError('Must implement `get_token` method for `TokenObtainSerializer` subclasses')


class TokenObtainPairSerializer(TokenObtainSerializer):
    @classmethod
    def get_token(cls, user, **kwargs):
        request = kwargs.pop('request')
        if api_settings.ADD_USER_IP_CLAIM:
            kwargs[api_settings.USER_IP_CLAIM] = get_real_ip_from_header_meta(request.META)
        if api_settings.ADD_USER_AGENT_CLAIM:
            kwargs[api_settings.USER_AGENT_CLAIM] = get_user_agent_from_header_meta(request.META)
        return RefreshToken.for_user(user, **kwargs)

    def validate(self, attrs):
        data = super().validate(attrs)
        request = self.context['request']
        refresh = self.get_token(self.user, request=request)

        data['refresh'] = str(refresh)
        data['access'] = str(refresh.access_token)

        return data


class VerificationBaseSerializer(serializers.Serializer):
    """
    Abstract serializer used for verifying and refreshing JWTs.
    """

    def validate(self, attrs):
        msg = 'Please define a validate method.'
        raise NotImplementedError(msg)

    def _check_user(self, refresh_token):
        user_id = refresh_token[api_settings.USER_ID_CLAIM]

        try:
            user = User.objects.get(pk=user_id)
        except User.DoesNotExist:
            msg = _("User doesn't exist.")
            raise InvalidUser(msg)

        if not user.is_active:
            msg = _('User account is disabled.')
            raise InvalidUser(msg)

        return user

class TokenRefreshSerializer(VerificationBaseSerializer):
    refresh = serializers.CharField()

    def validate(self, attrs):
        refresh = RefreshToken(attrs['refresh'])
        self._check_user(refresh)

        data = {'access': str(refresh.access_token)}

        if api_settings.ROTATE_REFRESH_TOKENS:
            refresh.set_jti()
            refresh.set_exp()

            data['refresh'] = str(refresh)

        return data


class TokenVerifySerializer(serializers.Serializer):
    token = serializers.CharField()

    def validate(self, attrs):
        UntypedToken(attrs['token'])

        return {}



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="djangomicroframework",
    version="0.0.2a1",
    license='MIT',
    author="thanhnb",
    author_email="bathanhtlu@gmail.com",
    install_requires=[
        'django>=2.0',
        'djangorestframework>=3.11',
        'pyjwt>=1.7',
        'redis>=3.3.11',
        'hiredis>=1.0.1'
    ],
    python_requires='>=3.6,<3.7',
    description="Micro framework for Django",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/lacoski/django-micro-framework",
    packages=setuptools.find_packages(exclude=['tests', 'tests.*', 'requirements']),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)
